!+
! Module: array_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module array_utils

  use dal
  use sscmisc

  implicit none 

  private 

  public valueAtBinFraction

  interface valueAtBinFraction
    module procedure valueAtBinFractionVector, valueAtBinFractionArray
  end interface

contains

  !---------------------------------------------------------------------
  function valueAtBinFractionVector(vector, binFraction)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: vector(:), binFraction
    real :: valueAtBinFractionVector

    integer :: nearestBin
    real :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    adjustedBinFraction = min(1.0, max(0.0, binFraction))
    nearestBin = max(1, ceiling(size(vector) * adjustedBinFraction))
    valueAtBinFractionVector = valueAtBinFractionKernel(vector, nearestBin)

  end function valueAtBinFractionVector

  !---------------------------------------------------------------------
  function valueAtBinFractionArray(array, binFraction, maskArray)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: array(:,:), binFraction
    logical, intent(in), optional :: maskArray(:,:)
    real :: valueAtBinFractionArray

    integer :: xsize, ysize, vsize, numMasked, nearestBin
!    real :: vector(size(array, 1) * size(array, 2))
real, allocatable :: vector(:)
!    real :: tempArray(size(array, 1), size(array, 2))
    real :: adjustedBinFraction
    logical :: doMasking

    !-------------------------------------------------------------------
    ! Start

    xsize = size(array, 1)
    ysize = size(array, 2)
    vsize = xsize * ysize

    doMasking = .false. ! default
    if (present(maskArray)) then
      numMasked = count(maskArray)

      if (xsize .eq. size(maskArray, 1)&
      .and. ysize .eq. size(maskArray, 2)&
      .and. numMasked .gt. 0) then
        doMasking = .true.
      end if
    end if

    if (doMasking) then
!      where(maskArray)
!        tempArray = array
!      elsewhere
!        tempArray = minval(array)
!      end where
!      vector = reshape(tempArray, (/vsize/))
allocate(vector(numMasked))
vector = pack(array, maskArray)

    else
      numMasked = vsize
allocate(vector(numMasked))
      vector = reshape(array, (/vsize/))
    end if

    adjustedBinFraction = min(1.0, max(0.0, binFraction))
    nearestBin = max(1, ceiling(numMasked * adjustedBinFraction))
!    nearestBin = nearestBin + (vsize - numMasked)

    valueAtBinFractionArray = valueAtBinFractionKernel(vector, nearestBin)

  end function valueAtBinFractionArray

  !---------------------------------------------------------------------
  function valueAtBinFractionKernel(vector, nearestBin)
  !---------------------------------------------------------------------
  ! Adapted from the Numerical Recipes function select().

!    use sscmisc

    implicit none

    real, intent(in) :: vector(:)
    integer, intent(in) :: nearestBin
    real :: valueAtBinFractionKernel

    integer :: vsize, i, ir, j, l, mid
    real :: tempVec(size(vector)), tempVal

    !-------------------------------------------------------------------
    ! Start

    vsize = size(vector)
    tempVec = vector

    l = 1
    ir = vsize

    do
      if (ir - l .le. 1) then
        if (ir - l .eq. 1) then
          if (tempVec(l) .gt. tempVec(ir)) call swap(tempVec(l), tempVec(ir))
        end if
        valueAtBinFractionKernel = tempVec(nearestBin)
    exit

      else
        mid = (l + ir) / 2
        call swap(tempVec(mid), tempVec(l + 1))

        if (tempVec(l)  .gt. tempVec(ir))  call swap(tempVec(l),   tempVec(ir))
        if (tempVec(l+1).gt. tempVec(ir))  call swap(tempVec(l+1), tempVec(ir))
        if (tempVec(l)  .gt. tempVec(l+1)) call swap(tempVec(l),   tempVec(l+1))

        i = l + 1
        j = ir
        tempVal = tempVec(l + 1)
        do
          do
            i = i + 1
          if (tempVec(i) .ge. tempVal) exit
          end do

          do
            j = j - 1
          if (tempVec(j) .le. tempVal) exit
          end do

        if (j .lt. i) exit

          call swap(tempVec(i), tempVec(j))
        end do

        tempVec(l + 1) = tempVec(j)
        tempVec(j) = tempVal
        if (j .ge. nearestBin) ir = j - 1
        if (j .le. nearestBin) l = i
      end if
    end do

  end function valueAtBinFractionKernel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module array_utils
