!
! Module: sscmisc
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: A grab-bag of general-purpose routines.
! 
! Comments: TODO:
! 
! $Id: sscmisc.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module sscmisc

  use dal
  use constants

  implicit none 

  private 

  public ucase
  public lcase
  public strToInt
  public intToStr
  public indicize
  public maskArrayLimits
  public isNaN
  public notNaN
  public calcMinPositive
  public calcMaxNegative
  public printEvery
  public int125
  public raDegToStr
  public decDegToStr
  public swap
  public getFreeIoUnit
  public fracVal

  interface fracVal
    module procedure fracValScalar, fracValVector
  end interface

  interface indicize
    module procedure indicize_int32
    module procedure indicize_int16
  end interface

  interface calcMinPositive
    module procedure calcMinPositive_real_vector
    module procedure calcMinPositive_real_array
  end interface

  interface calcMaxNegative
    module procedure calcMaxNegative_real_vector
    module procedure calcMaxNegative_real_array
  end interface

  interface raDegToStr
    module procedure raDegToStr_scalar
    module procedure raDegToStr_vector
  end interface

  interface decDegToStr
    module procedure decDegToStr_scalar
    module procedure decDegToStr_vector
  end interface

  logical :: firstCallPrintEvery = .true.
  integer :: myJump ! used by printEvery

contains 

  !---------------------------------------------------------------------
  real function fracValScalar(min, max, frac)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) :: min, max, frac

    !-------------------------------------------------------------------
    ! Start:

    fracValScalar = (1.0 - frac) * min + frac * max

  end function fracValScalar

  !---------------------------------------------------------------------
  function fracValVector(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) :: min, max, frac(:)
    real :: fracValues(size(frac))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + frac * max

  end function fracValVector

  !---------------------------------------------------------------------
  function ucase(arg) result(string)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: arg
    character(len=len(arg)) :: string
    integer :: i, asciiIndex, aAsciiIndex, zAsciiIndex, bigAasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    zAsciiIndex = ichar('z')
    bigAasciiIndex = ichar('A')
    string = arg
    do i = 1, len(arg)
      asciiIndex = ichar(string(i:i))
      if (asciiIndex .ge. aAsciiIndex .and. asciiIndex .le. zAsciiIndex) then
        string(i:i) = char(bigAasciiIndex + asciiIndex - aAsciiIndex)
      end if
    end do

  end function ucase

  !---------------------------------------------------------------------
  function lcase(arg) result(string)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: arg
    character(len=len(arg)) :: string
    integer :: i, asciiIndex, aAsciiIndex, bigAasciiIndex, bigZasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    bigAasciiIndex = ichar('A')
    bigZasciiIndex = ichar('Z')
    string = arg
    do i = 1, len(arg)
      asciiIndex = ichar(string(i:i))
      if (asciiIndex .ge. bigAasciiIndex&
      .and. asciiIndex .le. bigZasciiIndex) then
        string(i:i) = char(aAsciiIndex + asciiIndex - bigAasciiIndex)
      end if
    end do

  end function lcase

  !---------------------------------------------------------------------
  subroutine strToInt(str, int, status)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str
    integer, intent(out) :: int, status

    integer :: i, length, nonZeroLength, mult, digit
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testflag) print *, '        start of strToInt.'

    status = 0 ! default
    int    = 0 ! default

    length = len(str)
if (testflag) print *, '        len(input str) is ', length
if (testflag) print *, '        input str is ', str
    if (length .le. 0) then
      status = 1
  return
    end if

    do i = 1, length
      if (iachar(str(i:i)) .lt. iachar('0')&
      .or. iachar(str(i:i)) .gt. iachar('9')) then
        status = 2
  return
      end if
    end do

    nonZeroLength = 0
    do i = 1, length
      if (iachar(str(i:i)) .ne. iachar('0')) then
        nonZeroLength = length + 1 - i
    exit
      end if
    end do

if (testflag) print *, '        nonZeroLength is ', nonZeroLength
    if (nonZeroLength .eq. 0) then
      int = 0
  return
    end if

    if (nonZeroLength .gt. floor(log10(real(huge(int))))) then
    ! floor because eg huge(int16) is 32768, thus ceiling(log10(huge(int16)))
    ! is 5, allowing a str value of, for example, '32769', which would cause
    ! an overflow.
      status = 3
  return
    end if

    int = iachar(str(length:length)) - iachar('0')
if (testflag) print *, '        initial total (just = LS digit) is ', int
    mult = 1
if (testflag) then
  print *, '        stuff that controls end of loop:'
  print *, '          length=', length
  print *, '          nonZeroLength=', nonZeroLength
  print *, '          loop start is length - 1, =', length - 1
  print *, '          loop end is 1 + length - nonZeroLength, ='&
    , 1 + length - nonZeroLength
  print *, ''
end if
    do i = length - 1, 1 + length - nonZeroLength, -1
      mult = mult * 10
      digit = iachar(str(i:i)) - iachar('0')
      int = int + mult * digit
if (testflag) then
  print *, '          i=',i
  print *, '          mult=',mult
  print *, '          digit=',digit
  print *, '          total=',int
  print *, ''
end if
    end do

if (testflag) print *, '        end of strToInt.'
  end subroutine strToInt

  !---------------------------------------------------------------------
  function intToStr(int, leftJustify) result(str)
  !---------------------------------------------------------------------

    implicit none

    integer,      intent(in)  :: int
    logical, intent(in), optional  :: leftJustify
    character(len=(2+range(int))) :: str

    logical :: localLeftJustify

    !-------------------------------------------------------------------
    ! Start

    if (present(leftJustify)) then
      localLeftJustify = leftJustify
    else
      localLeftJustify = .true.
    end if

    str = '' ! default

    write(str, *) int
    if (localLeftJustify) then
      str = adjustl(str)
    else
      str = adjustr(str)
    end if

  end function intToStr

  !---------------------------------------------------------------------
  function indicize_int32(bound1, bound2) result(indexVector)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: bound1
    integer(int32), intent(in), optional :: bound2
    integer(int32), pointer :: indexVector(:)

    integer(int32) :: loi, hii, i

    !-------------------------------------------------------------------
    ! Start

    if (present(bound2)) then
      loi = bound1
      hii = bound2
    else
      loi = 1
      hii = bound1
    end if

    allocate(indexVector(loi:hii))

    do i = loi, hii
      indexVector(i) = i
    end do

  end function indicize_int32

  !---------------------------------------------------------------------
  function indicize_int16(bound1, bound2, testflag) result(indexVector)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: bound1
    integer(int16), intent(in), optional :: bound2
    integer(int16), pointer :: indexVector(:)
logical, intent(in), optional :: testflag

    integer(int16) :: loi, hii, i
logical :: localTestflag

    !-------------------------------------------------------------------
    ! Start

if (present(testflag)) then
localTestflag = testflag
else
localTestflag = .false.
end if

 if (localTestflag) print *, '      hello!'

    if (present(bound2)) then
      loi = bound1
      hii = bound2
 if (localTestflag) print *, '      loi, hii: ', loi, hii
    else
      loi = 1
      hii = bound1
    end if

    allocate(indexVector(loi:hii))
 if (localTestflag) print *, '      indexVector allocated sir!'

    do i = loi, hii
      indexVector(i) = i
 if (localTestflag) print *, '        i, indexVector(i): ', i, indexVector(i)
    end do

 if (localTestflag) print *, '      goodbye!'
  end function indicize_int16

  !---------------------------------------------------------------------
  subroutine maskArrayLimits(mask, minXi, maxXi, minYi, maxYi, status)
  !---------------------------------------------------------------------
  ! Adapted from CGP's im_lims subroutine in implot.

    implicit none

    logical, intent(in) :: mask(:,:)
    integer, intent(out) :: minXi, maxXi, minYi, maxYi
    integer, intent(out), optional :: status

    integer :: localStatus, xsize, ysize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    minXi = 0
    maxXi = 0
    minYi = 0
    maxYi = 0
    
    if (count(mask) .le. 0) then
      if (present(status)) status = -1
  return
    end if

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    if (xsize .le. 0 .or. ysize .le. 0) then
      if (present(status)) status = -2
  return
    end if

    do xi = 1, xsize
      if (any(mask(xi, :))) then
        minXi = xi
    exit
      end if
    end do
    do xi = xsize, 1, -1
      if (any(mask(xi, :))) then
        maxXi = xi
    exit
      end if
    end do
    do yi = 1, ysize
      if (any(mask(:, yi))) then
        minYi = yi
    exit
      end if
    end do
    do yi = ysize, 1, -1
      if (any(mask(:, yi))) then
        maxYi = yi
    exit
      end if
    end do

  end subroutine maskArrayLimits

  !---------------------------------------------------------------------
  logical function isNaN(arg)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: arg

    !-------------------------------------------------------------------
    ! Start

    if (notNaN(arg)) then
      isNan = .false.
    else
      isNan = .true.
    end if

  end function isNaN

  !---------------------------------------------------------------------
  logical function notNaN(arg)
  !---------------------------------------------------------------------
  ! Fairly fudgy. Don't rely on it, as definition of NaN may be different on other platforms?

    implicit none

    real(single), intent(in) :: arg

    integer(int32) :: myInt32 = 1

    !-------------------------------------------------------------------
    ! Start

    if (transfer(arg, myInt32) .eq. huge(myInt32)) then
      notNaN = .false.
    else
      notNaN = .true.
    end if

  end function notNaN

  !---------------------------------------------------------------------
  subroutine calcMinPositive_real_vector(vector, minPositiveValue&
  , allNonPositive)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: vector(:)
    real, intent(out) :: minPositiveValue
    logical, intent(out) :: allNonPositive

    real :: array(1,size(vector))

    !-------------------------------------------------------------------
    ! Start

    array(1,:) = vector
    call calcMinPositive_real_array(array, minPositiveValue, allNonPositive)

  end subroutine calcMinPositive_real_vector

  !---------------------------------------------------------------------
  subroutine calcMinPositive_real_array(array, minPositiveValue&
  , allNonPositive)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: array(:,:)
    real, intent(out) :: minPositiveValue
    logical, intent(out) :: allNonPositive

    integer :: xsize, xi, startXi, ysize, yi, startYi

    !-------------------------------------------------------------------
    ! Start

    if (all(array .le. 0.0)) then
      minPositiveValue = 0.0 ! just to give it a value.
      allNonPositive = .true.
  return
    end if

    allNonPositive = .false.
    xsize = size(array, 1)
    ysize = size(array, 2)

    XLOOP: do xi = 1, xsize
    if (all(array(xi,:) .le. 0.0)) cycle
      do yi = 1, ysize
        if (array(xi,yi) .gt. 0.0) then
          startXi = xi
          startYi = yi
          minPositiveValue = array(xi,yi) ! starting value.
    exit XLOOP
        end if
      end do
    end do XLOOP

    do xi = startXi, xsize
    if (all(array(xi,:) .le. 0.0)) cycle
      do yi = startYi + 1, ysize
        if (array(xi,yi) .gt. 0.0) then
          minPositiveValue = min(minPositiveValue, array(xi,yi))
        end if
      end do
    end do

  end subroutine calcMinPositive_real_array

  !---------------------------------------------------------------------
  subroutine calcMaxNegative_real_vector(vector, maxNegativeValue&
  , allNonNegative)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: vector(:)
    real, intent(out) :: maxNegativeValue
    logical, intent(out) :: allNonNegative

    real :: array(1,size(vector))

    !-------------------------------------------------------------------
    ! Start

    array(1,:) = vector
    call calcMaxNegative_real_array(array, maxNegativeValue, allNonNegative)

  end subroutine calcMaxNegative_real_vector

  !---------------------------------------------------------------------
  subroutine calcMaxNegative_real_array(array, maxNegativeValue&
  , allNonNegative)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: array(:,:)
    real, intent(out) :: maxNegativeValue
    logical, intent(out) :: allNonNegative

    integer :: xsize, xi, startXi, ysize, yi, startYi

    !-------------------------------------------------------------------
    ! Start

    if (all(array .ge. 0.0)) then
      maxNegativeValue = 0.0 ! just to give it a value.
      allNonNegative = .true.
  return
    end if

    allNonNegative = .false.
    xsize = size(array, 1)
    ysize = size(array, 2)

    XLOOP: do xi = 1, xsize
    if (all(array(xi,:) .ge. 0.0)) cycle
      do yi = 1, ysize
        if (array(xi,yi) .lt. 0.0) then
          startXi = xi
          startYi = yi
          maxNegativeValue = array(xi,yi) ! starting value.
    exit XLOOP
        end if
      end do
    end do XLOOP

    do xi = startXi, xsize
    if (all(array(xi,:) .ge. 0.0)) cycle
      do yi = startYi + 1, ysize
        if (array(xi,yi) .lt. 0.0) then
          maxNegativeValue = max(maxNegativeValue, array(xi,yi))
        end if
      end do
    end do

  end subroutine calcMaxNegative_real_array

  !---------------------------------------------------------------------
  subroutine printEvery(i, total, jump, fraction, message)
  !---------------------------------------------------------------------
  ! This is designed for monitoring the progress of large loops. In this
  ! situation a print statement that is called at every loop pass can
  ! generate a large amount of output and slow execution down a lot. The
  ! present subroutine is designed to print a message only every so often.
  ! The arguments 'jump' and 'fraction' are optional. If 'jump' is
  ! specified, the message is printed every 'jump' cycles. If 'jump' is
  ! not supplied but 'fraction' is, the jump number is calculated from
  ! fraction * total. If neither is given, the message is printed at
  ! every cycle.

    implicit none

    integer,      intent(in)           :: i, total
    integer,      intent(in), optional :: jump
    real,         intent(in), optional :: fraction
    character(*), intent(in), optional :: message

    integer :: percent

    !-------------------------------------------------------------------
    ! Start

    if (firstCallPrintEvery .and. present(fraction)) then
      firstCallPrintEvery = .false.
      myJump = min(total, max(1, nint(fraction * total)))
    end if

    if (present(jump)) then
      if (modulo(i, jump) .eq. 0) then
        if (present(message)) then
          print *, message, i, '/', total
        else
          print *, i, '/', total
        end if
      end if
    else if (present(fraction)) then
      if (modulo(i, myJump) .eq. 0) then
        percent = nint(100 * ((i - 1) / real(total - 1)))
        if (present(message)) then
          print *, message, percent, '%'
        else
          print *, percent, '%'
        end if
      else if (i >= total) then
        firstCallPrintEvery = .true.
      end if
    else
      if (present(message)) then
        print *, message, i, '/', total
      else
        print *, i, '/', total
      end if
    end if

  end subroutine printEvery

  !---------------------------------------------------------------------
  subroutine printEvery_old(message, i, total, jump, fraction)
  !---------------------------------------------------------------------
  ! This is designed for monitoring the progress of large loops. In this situation a print statement that is called at every loop pass can generate a large amount of output and slow execution down a lot. The present subroutine is designed to print a message only every so often. The arguments 'jump' and 'fraction' are optional. If 'jump' is specified, the message is printed every 'jump' cycles, starting at the first cycle. If 'jump' is not supplied but 'fraction' is, the jump number is calculated from fraction * total, such that the calculated jump is the nearest integer of the type 1, 2, 5 etc. If neither is given, the message is printed at every cycle.

    implicit none

    character(*), intent(in)           :: message
    integer,      intent(in)           :: i, total
    integer,      intent(in), optional :: jump
    real,         intent(in), optional :: fraction

    integer :: myJump

    !-------------------------------------------------------------------
    ! Start

    if (present(jump)) then
      if (i .eq. 1 .or. modulo(i, jump) .eq. 0) print *, message, i, '/', total
    else if (present(fraction)) then
      myJump = int125(fraction * total)
      if (i .eq. 1 .or. modulo(i,myJump).eq. 0) print *, message, i, '/', total
    else
      print *, message, i, '/', total
    end if

  end subroutine printEvery_old

  !---------------------------------------------------------------------
  integer function int125(arg)
  !---------------------------------------------------------------------
  ! This function finds the value of 2, 5 or 10 which is nearest (in a
  ! logarithmic sense) to the input argument.

    implicit none

    real, intent(in) :: arg

    integer :: sign, mantissa
    real :: log10Arg, logRemainder, localArg

    !-------------------------------------------------------------------
    ! Start

    if (arg .eq. 0.0) then
      int125 = 0
  return
    end if

    if (arg .lt. 0.0) then
      sign = -1
      localArg = -arg
    else
      sign = 1
      localArg = arg
    end if

    log10Arg = log10(localArg)
    mantissa = floor(log10Arg)
!**** test mantissa to guard against integer overflows.

    logRemainder = log10Arg - mantissa
    if (logRemainder .lt. lnTwo / lnTen / 2.0) then ! less than 0.150515,
    ! ie half-way between log10(1) and log10(2).

      int125 = 10**mantissa

    elseif (logRemainder .lt. 0.5) then ! less than 0.500000,
    ! ie half-way between log10(2) and log10(5).

      int125 = 2 * 10.0**mantissa

    elseif (logRemainder .lt. 1.0 - lnTwo / lnTen / 2.0) then ! less than
    ! 0.849485, ie half-way between log10(5) and log10(10).

      int125 = 5 * 10.0**mantissa

    else
      int125 = 10**(mantissa + 1)
    end if

    int125 = int125 * sign

  end function int125

  !---------------------------------------------------------------------
  subroutine raDegToStr_scalar(raDeg,  raStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: raDeg
    character(11), intent(out) :: raStr

    real(double) :: raDegVec(1)
    character(11) :: raStrVec(1)

    !-------------------------------------------------------------------
    ! Start

    raDegVec(1) = raDeg
    call raDegToStr_vector(raDegVec, raStrVec)
    raStr = raStrVec(1)

  end subroutine raDegToStr_scalar

  !---------------------------------------------------------------------
  subroutine raDegToStr_vector(raDeg,  raStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: raDeg(:)
    character(11), intent(out) :: raStr(size(raDeg))

    real(single) :: raHours, raMin, raSec
    integer(int16) :: raHoursInt, raMinInt, i
    character(5) :: raSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(raDeg)
      raHours = raDeg(i) / 15.0
      raHoursInt = floor(raHours)
      raMin = (raHours - raHoursInt) * 60.0
      raMinInt = floor(raMin)
      raSec = (raMin - raMinInt) * 60.0

      if (raSec .lt. 1.0) then
        write(raSecStr, '(a2, f3.2)') '00', raSec
      else if (raSec .lt. 10.0) then
        write(raSecStr, '(a1, f4.2)') '0', raSec
      else
        write(raSecStr, '(f5.2)') raSec
      end if

      write(raStr(i), '(i2.2, a1, i2.2, a1, a5)') raHoursInt, ':', raMinInt&
      , ':', raSecStr
    end do

  end subroutine raDegToStr_vector

  !---------------------------------------------------------------------
  subroutine decDegToStr_scalar(decDeg,  decStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: decDeg
    character(12), intent(out) :: decStr

    real(double) :: decDegVec(1)
    character(12) :: decStrVec(1)

    !-------------------------------------------------------------------
    ! Start

    decDegVec(1) = decDeg
    call decDegToStr_vector(decDegVec, decStrVec)
    decStr = decStrVec(1)

  end subroutine decDegToStr_scalar

  !---------------------------------------------------------------------
  subroutine decDegToStr_vector(decDeg, decStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: decDeg(:)
    character(12), intent(out) :: decStr(size(decDeg))

    real(single) :: absDecDeg, decMin, decSec
    integer(int16) :: absDecDegInt, decMinInt, i
    character(1) :: decSignStr
    character(5) :: decSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(decDeg)
      if(decDeg(i) .lt. 0.d0)then
        decSignStr = '-'
      else
        decSignStr = '+'
      endif

      absDecDeg = abs(decDeg(i))
      absDecDegInt = floor(absDecDeg)
      decMin = (absDecDeg - absDecDegInt) * 60.0
      decMinInt = floor(decMin)
      decSec = (decMin - decMinInt) * 60.0

      if (decSec .lt. 1.0) then
        write(decSecStr, '(a2, f3.2)') '00', decSec
      else if (decSec .lt. 10.0) then
        write(decSecStr, '(a1, f4.2)') '0', decSec
      else
        write(decSecStr, '(f5.2)') decSec
      end if

!print *, 'decSignStr: ', decSignStr
!print *, 'absDecDegInt: ', absDecDegInt
!print *, 'decMinInt: ', decMinInt
!print *, 'decSec: ', decSec
!print *, 'decSecStr: ', decSecStr
!print *, ''
      write(decStr(i), '(a1, i2.2, a1, i2.2, a1, a5)') decSignStr, absDecDegInt&
      , ':', decMinInt, ':', decSecStr
    end do

  end subroutine decDegToStr_vector

  !---------------------------------------------------------------------
  subroutine swap(a, b)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) :: a, b
    real :: temp

    !-------------------------------------------------------------------
    ! Start

    temp = a
    a = b
    b = temp

  end subroutine swap

  !---------------------------------------------------------------------
  subroutine getFreeIoUnit(unitNum)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(out) :: unitNum   ! unit number free for use.
    logical :: exists, open

    !-------------------------------------------------------------------
    ! Start

    do unitNum = 99, 0, -1
       inquire(unit=unitNum, exist=exists, opened=open)
  if (exists .and. .not. open) return
    end do

  end subroutine getFreeIoUnit

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module sscmisc
